@import url("css/theme.css");

.toggle .header {
    display: block;
    clear: both;
    padding-top: -10px;
    padding-bottom: 20px;
}

.toggle .header:after {
    content: " ▼";
}

.toggle .header.open:after {
    content: " ▲";
}

.wy-dropdown > a img.logo { width: 300px; }

/* Custom colors */

.wy-side-nav-search, .wy-nav-top { /* above navigation, behind logo */
    background: #51557E;
}

.icon.icon-home { /* link above logo */
    color: #D6D5A8;
}

.version { /* version spec under logo */
    color: #D6D5A8 !important;
}

.highlight { /* code blocks */
    background: #fff7e9;
}

.admonition-title { /* notes, title */
    background: #816797 !important;
}

.admonition { /* notes */
    background: #51557E !important;
}

.rst-content dl:not(.docutils) dt { /* class definitions */
    background: #e4f3f2;
    color: #005248;
    border-top: solid 3px #005248;
}

.rst-content dl:not(.docutils) dl dt{ /* functions */
    border-left:solid 3px #005248;
    background: #e4f3f2;
    color: #005248;
}


/* For alert boxes in jupyter notebooks */
.alert {
  padding: 8px;
  margin-bottom: 8px;
  border: 1px solid transparent;
  border-radius: 4px;
}
.alert h4 {
  margin-top: 0;
  color: inherit;
}
.alert .alert-link {
  font-weight: bold;
}
.alert > p,
.alert > ul {
  margin-bottom: 0;
}
.alert > p + p {
  margin-top: 5px;
}
.alert-dismissable,
.alert-dismissible {
  padding-right: 35px;
}
.alert-dismissable .close,
.alert-dismissible .close {
  position: relative;
  top: -2px;
  right: -21px;
  color: inherit;
}
.alert-success {
  background-color: #dff0d8;
  border-color: #d6e9c6;
  color: #3c763d;
}
.alert-success hr {
  border-top-color: #c9e2b3;
}
.alert-success .alert-link {
  color: #2b542c;
}
.alert-info {
  background-color: #d9edf7;
  border-color: #bce8f1;
  color: #31708f;
}
.alert-info hr {
  border-top-color: #a6e1ec;
}
.alert-info .alert-link {
  color: #245269;
}
.alert-warning {
  background-color: #fcf8e3;
  border-color: #faebcc;
  color: #8a6d3b;
}
.alert-warning hr {
  border-top-color: #f7e1b5;
}
.alert-warning .alert-link {
  color: #66512c;
}
.alert-danger {
  background-color: #f2dede;
  border-color: #ebccd1;
  color: #a94442;
}
.alert-danger hr {
  border-top-color: #e4b9c0;
}
.alert-danger .alert-link {
  color: #843534;
}
